/*
 *  seed.cc created on Mon, 31 May 1999 21:12:51 -0400 by noe
 */    
#ifndef __SEED_H__
#define __SEED_H__

//STL
#include <vector>
#include <queue>
#include <stack>
#include <list>
//STD
#include <assert.h>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
using namespace std;
#include "macro.h"

#ifdef SIGNATURE_PRECOMPUTED
/**
 * @brief node structure represents a tree node
 */

 typedef struct _node {
   vector<struct _node> * nodes;
   bool   min_reached;
   double pr;
 } node;
#endif
   
/**
 * @class seed
 * @brief represents a subset seed 
 */

class seed {
 public:

  /** @name  seed(int span) 
   *  @brief build a seed of given span but full of jokers
   *  @param span is the lenght of the seed
   */
 
  seed(int span):_span(span){

#ifdef SIGNATURE_PRECOMPUTED
    if(!_signature_prob)
      _signature_prob = set_signature_prob();
#endif

    _seedMotif_int  = new int [_span];
    _swap_from = 0;
    _swap_to   = 0;

    for(int i=0; i<_span; i++)
      _seedMotif_int[i] = 0;

    if (gv_signature_flag){
      checksignature();
      setsignature();
    }

    while(!acceptable())
      next();
  }


  /** @name  seed() 
   *  @brief build a seed of gv_minspan but full of jokers
   */

  seed():_span(gv_minspan){    
#ifdef SIGNATURE_PRECOMPUTED
    if(!_signature_prob)
      _signature_prob = set_signature_prob();
#endif
    _seedMotif_int  = new int [_span];
    _swap_from = 0;
    _swap_to   = 0;
    for(int i=0; i<_span; i++)
      _seedMotif_int[i] = 0;

    if (gv_signature_flag){
      checksignature();
      setsignature();
    }

    while(!acceptable())
      next();
  }




  /** @name  seed(char * desc) 
   *  @brief build a seed according to a descriptor 
   *  @param desc is a string in ['0'-'9''A'-'Z''a'-'z']
   */
  
  //@{
  
  seed (char * desc) {
#ifdef SIGNATURE_PRECOMPUTED
    if(!_signature_prob)
      _signature_prob = set_signature_prob();
#endif
    /* seed span */
    _span = 0;
    _swap_from = 0;
    _swap_to   = 0;

    for(char * pc = desc; *pc ; pc++)
      _span++;

    /* seed motif */
    _seedMotif_int  = new int [_span];
    for(int i = 0; i < _span ; i++)
      _seedMotif_int[i] = convert(desc[i]);
  }
  
  
  seed(string & str_desc) {
#ifdef SIGNATURE_PRECOMPUTED
    if(!_signature_prob)
      _signature_prob = set_signature_prob();
#endif
    const char * desc = str_desc.c_str();
    /* seed span */
    _span = 0;
    _swap_from = 0;
    _swap_to   = 0;
    for(char * pc = (char *) desc; *pc ; pc++)
      _span++;

    /* seed motif */
    _seedMotif_int  = new int [_span];
    for(int i = 0; i < _span ; i++)
      _seedMotif_int[i] = convert(desc[i]);
  }
  
  //@}

  
  ~seed(){
    if  (_seedMotif_int){
      delete _seedMotif_int;
      _seedMotif_int = NULL;
    }
  };


  /** @name Get functions
   *  @brief these methods are provided to get information from current seed 
   */
  
  //@{

  /** @brief gives the integer representation of the seed
   *  @return an integer table of lenght span, each element must 
   *          be in the range 0 .. |B|-1
   */
  
  int  *   table()   {return _seedMotif_int;};
  
  /** @brief gives the span of the seed
   *  @return the span of the seed
   */
  int      span()    {return _span;};
  
        
  /** @brief gives the current seed representation (a string
   *  @brief in ['0'-'9''A'-'Z''a'-'z'])
   *  @return a string object of length span
   */
  string str(){
    string result("");
    for (int i = 0 ; i < _span ; i++) {
      result += rconvert(_seedMotif_int[i]);
    }
    return result;
  }
  
  
  //@}

  
  /** @brief compute the weight of a given seed
   *  @return the weight
   */
  double weight(){
    double weight = 0.0;
    for(int i=0; i<_span; i++)
      weight += gv_bsel_weight[_seedMotif_int[i]];
    return weight;
  }
  

  /** @brief check if the current seed is acceptable : correct weight  (according to gloval variables gv_minweight and gv_maxweight)
   *  @return 1 if the current seed is acceptable
   */
  bool acceptable() {
    // FIXME : it is necessary to keep positive weighted elements on the seed borders ??

    if (!gv_vectorized && (
	gv_bsel_weight[_seedMotif_int[0]]       <= (1e-32) || 
	gv_bsel_weight[_seedMotif_int[_span-1]] <= (1e-32) )) {
      return false;
       }
    
    double w = weight();
    
    if (gv_weight_interval && (w < gv_minweight || w > gv_maxweight))
      return false;
    
    return true;
  }
  

  /** @brief check if the signature is realisable according to the span : stop the program otherwise
   */
  
  void checksignature() {
    int sumsignature = 0;
    for(int i = 0 ; i < gv_seed_alphabet_size ; i++)
      sumsignature += gv_signature[i];      
    if (sumsignature > gv_minspan ){
      ERROR("Invalid Signature","size of the signature is " << sumsignature << " and does not fit inside minimal span " << gv_minspan);
    }
  }

  /** @brief set the very first seed signature
   */

  void setsignature() {
    int pos = 0;
    for(int b = gv_seed_alphabet_size-1 ; b >= 0 && pos < _span ; b--){
      int lettersignature =  gv_signature[b];
      while(lettersignature > 0) { // fill with b
	_seedMotif_int[pos++] = b;
	lettersignature--;
      }
    }
    while (pos < _span) { // fill remaining part with jokers
      _seedMotif_int[pos++] = 0;
    }
  }



  /** @brief set all the elements of position >= first_pos that are <= b in the seed in the decreasing order
   *  @brief this function is used to enumerate signatures and inner used.
   *  @param first_pos gives the first position to be sorted
   *  @param b gives the maximal value that can be sorted (all others remain unchanged)
   *  @see next()
   */

  void reorder(int first_pos,int b) {
    int last_bm = -1;
    for(int i = first_pos ; i < _span ; i++) {
      if (_seedMotif_int[i] == b && last_bm >= 0) {
	//swap elements
	int e = _seedMotif_int[i];
	_seedMotif_int[i] = _seedMotif_int[last_bm];
	_seedMotif_int[last_bm] = e;
	i = last_bm - 1;
	last_bm = -1;
      } else {
	if (_seedMotif_int[i] < b && last_bm == -1)
	  last_bm = i;
      }
    }
  }


  /** @brief generates the next seed 
   *  @brief this function does enumerative generation of acceptable seeds, with ou without signature constraits
   *  @return 1 if next seed has been generated, 0 if all the seeds in the set have already been generated
   *  @see acceptable()
   *  @see setsignature(),checksignature()
   */

  int next() {
    if (gv_signature_flag) {
      
    start_sign:
      /* 1) non trivial signature enumeration */
      for (int b = 1; b < gv_seed_alphabet_size ; b++ ){
	
	int first_b_from_right_to_left = -1;
	int last_bm_before_b           = -1; 
	
	for (int j=_span-1; j >= 0; j--) {
	  if ( _seedMotif_int[j] <  b ) {
	    last_bm_before_b = j; 
	  } else if ( _seedMotif_int[j] == b && last_bm_before_b >= 0 ) {
	    first_b_from_right_to_left = j ;
	    break;
	  }
	}
	if (first_b_from_right_to_left >= 0) {
	  //swap elements 
	  {
	    int e = b;
	    _seedMotif_int[first_b_from_right_to_left] =  _seedMotif_int[last_bm_before_b];
	    _seedMotif_int[last_bm_before_b] = e;
	  }
	  //reorder elements b and <b after this swap position
	  reorder(last_bm_before_b,b);

	  //reorder the elements <b (only at position where such elements are)
	  for (int bl = b-1 ; bl >= 0 ; bl-- )
	    reorder(0,bl);
	  while(!acceptable())
	    goto start_sign;
	  return 1;
	}
      }
      
      // span changing
      if (_span < gv_maxspan) {
	_span ++;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
	setsignature();
	while(!acceptable())
	  goto start_sign;
	return 1;
      } else {
	return 0;
      }


    } else {
      
      /* 2) complete enumeration */
    start:
      int j=_span-1;
      
      while(_seedMotif_int[j] == (gv_seed_alphabet_size-1)){
	j--;
	if (j < 0) {
	  if (_span < gv_maxspan) {
	    _span ++;
	    delete _seedMotif_int;
	    _seedMotif_int = new int[_span];
	    goto reset;
	  } else {
	    return 0;
	  }
	}
      }
      
      _seedMotif_int[j]++;
    reset:
      j++;
      for(int i = j ; i < _span ; i++) {
	_seedMotif_int[i] = 0;
      }
      
      while(!acceptable()){
	goto start;
      }
      return 1;
    }/* 2) */
    
  }
  
  
  


  /** @brief generates a random seed according to gv_minspan and gv_maxspan 
   *  @return 1 all the time
   */
  int random(){
    if (gv_signature_flag) {


      /* 1) signature random generation */
      // check signature
      int sumsignature = 0;
      for(int i = 0 ; i < gv_seed_alphabet_size ; i++)
	sumsignature += gv_signature[i];      
      if (sumsignature > gv_minspan ){
	ERROR("Invalid Signature","size of the signature is " << sumsignature << " and does not fit inside minimal span " << gv_minspan);
      }
      
      // select a new span 
      int newspan = rand()%(gv_maxspan - gv_minspan + 1) + gv_minspan;
      if (newspan != _span) {
	_span = newspan;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
      }
       
      
      // "fill and swap" randomization : 

      // fill
      setsignature();      
      // swap
    swap:
      for(int i= 0 ; i < _span ; i++){
	int j = rand()%_span;
	int letter_tmp    = _seedMotif_int[i];
	_seedMotif_int[i] = _seedMotif_int[j];
	_seedMotif_int[j] = letter_tmp;
      }
      
     
      while(!acceptable())
	goto swap; /* FIXME  can loop infinitely if a bad signature is given */

    } else {

    start:
      /* 2) complete enumeration */
      int newspan = rand()%(gv_maxspan - gv_minspan + 1) + gv_minspan;
      if (newspan != _span) {
	_span = newspan;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
      }
      for(int i = 0 ; i < _span ; i++) {
	_seedMotif_int[i] = rand() % gv_seed_alphabet_size;
      }



      /* heuristic weight trick to modify the seed to reach the reasonable weight */
      if (gv_weight_interval) {
	
	double w   = weight();
	
	double delta_min = w - gv_minweight;
	double delta_max = w - gv_maxweight;	
	// randomly change unty we reach [-0.5 .. +0.5]
	if (delta_min < -0.5 || delta_max > 0.5) {
	  for(int i = 0 ; i < _span ; i++) {
	    while(_seedMotif_int[i] < gv_seed_alphabet_size - 1 && delta_min < -0.5) {
	      int      bold = _seedMotif_int[i]++;
	      double  delta = gv_bsel_weight[_seedMotif_int[i]] - gv_bsel_weight[bold];
	      w         += delta;
	      delta_min += delta;
	      delta_max += delta;
	    }
	    while(_seedMotif_int[i] > 0                         && delta_max >   0.5) {
	      int      bold = _seedMotif_int[i]--;
	      double  delta = gv_bsel_weight[_seedMotif_int[i]] - gv_bsel_weight[bold];
	      w         += delta;
	      delta_min += delta;
	      delta_max += delta;
	    }
	  }	    
	}
	
	if (delta_min >= -0.5 && delta_max <= 0.5) {	  
	  for(int i = 0 ; i < _span ; i++) { // for each seed element
	    double w_i = gv_bsel_weight[_seedMotif_int[i]];	  
	    for(int b = 0 ; b < gv_seed_alphabet_size ; b++ ) { // for each seed letter
	      double w_b = gv_bsel_weight[b];
	      double   t = w_i - w_b; // compute difference
	      if (delta_min < 0 && t > delta_min && t < delta_max || delta_max < 0 && t < delta_min && t > delta_max  ) {
		_seedMotif_int[i] = b;
		goto eol;
	      } 
	    }
	  }
	}
      eol:;
      }
      
      while(!acceptable()){
	goto start; 
      }      
    }
    return 1;
  }


  /** @brief does a swap for the hill climbing function
   *  @return 1 if the next swap has been generated, 0 if all swap have been finished
   */
  int nextswap() {

    /* reverse previous swap */
    {
      int letter_tmp             = _seedMotif_int[_swap_from];
      _seedMotif_int[_swap_from] = _seedMotif_int[_swap_to];
      _seedMotif_int[_swap_to]   = letter_tmp;
    }

    /* move _swap...  positions */
    do {
      _swap_to++;
      if (_swap_to >= _span) {
	_swap_from++;    
	if (_swap_from >= _span - 1) {
	  _swap_to = _swap_from = 0;    
	  return 0;
	}
	_swap_to = _swap_from + 1;
      }
    } while(_seedMotif_int[_swap_from] == _seedMotif_int[_swap_to]);
    
    /* swap */
    {
      int letter_tmp             = _seedMotif_int[_swap_from];
      _seedMotif_int[_swap_from] = _seedMotif_int[_swap_to];
      _seedMotif_int[_swap_to]   = letter_tmp;
    }    
    return 1;
  }


  
  /** @brief reset the swap for the hill climbing function, all swap will then be processed from the begining
   */
  void resetswap() {
    _swap_from = 0;
    _swap_to   = 0;
  }


  /** @brief fix definitely the swap
   */
  void setswap() {
    /* swap */
    {
      int letter_tmp             = _seedMotif_int[_swap_from];
      _seedMotif_int[_swap_from] = _seedMotif_int[_swap_to];
      _seedMotif_int[_swap_to]   = letter_tmp;
    }  
    _swap_from = 0;
    _swap_to   = 0;
  }











  
  /** @brief seed ouput operator
   *  @return outputstream
   */
  
  friend ostream& operator<<(ostream& os, seed& s);



#ifdef SIGNATURE_PRECOMPUTED
   
  /** @brief compute seed probability to get a given signature
   *  @return a tree of possible signatures
   * 
   *
   * letter : 
   *     0     1          2
   *   0[x]  
   *   1[ ] ----->[f]--->[ ]
   *   2[ ] ->[f] [f]
   *              [f]
   *
   */
 
  static  vector<node> * set_signature_prob(int    letter     = gv_seed_alphabet_size-1, 
					    double min_weight = gv_minweight,
					    double max_weight = gv_maxweight,
					    int    min_span   = gv_minspan,
					    int    max_span   = gv_maxspan
					    ) {
    double element_weight     = gv_bsel_weight[letter];
    int    upper_bound_letter = (int)MIN(max_weight/element_weight,max_span);
    if (!seed::_signature)
      seed::_signature = new vector<int>(gv_seed_alphabet_size,-1);
    
    if (!_signature_prob && gv_weight_interval 
	&& max_weight > 0 
	&& element_weight > 0 
	&& upper_bound_letter > 0 
	&& letter >= 0) { 
      /*
      for(int i = letter  ; i <  gv_seed_alphabet_size-1 ;i++)
	cerr << "   " ;
      cerr << "( letter : " << letter << ", table size : " << upper_bound_letter  << " weights : [" << min_weight << "," << max_weight << "]" << endl;
      */
      vector<node> * table = new vector<node>(upper_bound_letter+1);
      
      /* recursively build the table */
      bool   empty = true;

      for( int a = 0 ; a <=  upper_bound_letter ; a++) {
	/* set the signature */
	(*seed::_signature)[letter] = a;
	
	(*table)[a].nodes =  set_signature_prob(letter-1,
						min_weight,
						max_weight,
						min_span,
						max_span
						);
	(*table)[a].min_reached = min_weight <= 0 && min_span <= 0;


	if((*table)[a].min_reached){
	  for(int i = gv_seed_alphabet_size-1 ; i >= a ; i--)
	    cerr << " s[" << i << "]=" << (*seed::_signature)[i] << "," ;
	  cerr << endl;
	}

	  
	min_weight -= element_weight;
	max_weight -= element_weight;
	min_span--;
	max_span--;
	empty  = empty  && ((*table)[a].nodes == NULL && (*table)[a].min_reached == false);
      }
      /*
      for(int i = letter  ; i <  gv_seed_alphabet_size-1 ;i++)
	cerr << "   " ;
      cerr << ") empty : " << empty << endl;
      */

      /* if none of the children has a solution */
      if (empty) {
	table->clear();
	delete table;	
	table = NULL;
      }

      return table;
    }
    return NULL;
  }
    
#endif


 protected:
  /// Current seed motif represented as an int array {0,1,2, ...} of lenght "_span"
  int  * _seedMotif_int;
  /// Fixed seed span 
  int    _span;
  /// Swapped char positions
  int    _swap_from,_swap_to;
#ifdef SIGNATURE_PRECOMPUTED
  /// Signature table probability
  static vector<node> * _signature_prob;
  static vector<int>  * _signature;
#endif

  /// convert a char into a number
  /// '0' -> 0, '1' -> 1 ... '9' -> 9, 'A' -> 10 ... 'Z' -> 35, 'a' -> 36 ...
  int convert(char c){
    if (c >= '0' && c <= '9') {
      if (c - '0' >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char \'" << c << "\' inside the seed descriptor" << endl;
	exit(-1);
      }
      return c - '0';
    } else if (c >= 'A' && c <= 'Z') {
      if (c - 'A' + 10 >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char \'" << c << "\' inside the seed descriptor" << endl;
	exit(-1);
      }
      return c - 'A' + 10;
    } else if (c >= 'a' && c <= '~') {
      if (c - 'a' + 36 >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char \'" << c << "\' inside the seed descriptor" << endl;
	exit(-1);
      } 
      return c - 'a' + 36;
    } else if (c >= '!') {
      if (c - '!' + 66 >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char \'" << c << "\' inside the seed descriptor" << endl;
	exit(-1);
      } 
      return c - '!' + 66;
    } else {
      cerr << "* Error : out of bound char \'" << c << "\' inside the seed descriptor" << endl;
      exit(-1);
      return 0;
    }
    return 0;
  }
  
  /// convert a number into a char
  char rconvert(int i){
    if (i < 10)
      return (char)i + '0';
    else 
      if (i < 36)
	return (char)i - 10 + 'A';
      else
	if (i < 66)
	  return (char)i - 36 + 'a';
	else
	  return (char)i - 66 + '!'; 
  }
 
};

#endif

